import os
import platform
import re
import subprocess
from distutils.version import LooseVersion

import sys

from Cython.Build.Dependencies import default_create_extension
from setuptools import setup
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext
from Cython.Build import cythonize


class CMakeExtension(Extension):
    def __init__(self, name, sources, cmake_options, *args, **kw):
        super().__init__(name, sources, *args, **kw)
        cmake_options["dir"] = os.path.abspath(cmake_options["dir"])
        self.cmake_options = cmake_options


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.run(['cmake', '--version'],
                                 stdout=subprocess.PIPE, check=True,
                                 universal_newlines=True)
        except (FileNotFoundError, subprocess.CalledProcessError):
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.stdout).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        super().run()

    def build_extension(self, ext):
        cmake_options = ext.cmake_options
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_ARCHIVE_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable,
                      '-DCMAKE_POSITION_INDEPENDENT_CODE=YES']
        library_output_dir = extdir

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_ARCHIVE_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2 ** 32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m', '/verbosity:minimal']
            library_name_format = "{}.lib"
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']
            library_name_format = "lib{}.a"

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.run(
            ['cmake', cmake_options["dir"]] + cmake_args,
            cwd=self.build_temp,
            env=env,
            check=True
        )
        if self.force:
            subprocess.run(
                ['cmake', '--build', '.', '--target', 'clean'] + build_args,
                cwd=self.build_temp,
                check=True
            )

        for target, target_output in cmake_options["targets"].items():
            subprocess.run(
                ['cmake', '--build', '.', '--target', target] + build_args,
                cwd=self.build_temp,
                check=True
            )
            ext.extra_objects.append(
                os.path.join(library_output_dir, library_name_format.format(target))
            )

        super().build_extension(ext)


is_windows = True

nod_submodule = os.path.join(os.path.dirname(__file__), "external", "nod")

custom_include_paths = [
    os.path.join(nod_submodule, "include"),
    os.path.join(nod_submodule, "logvisor", "include"),
]

extra_compile_args = []

if is_windows:
    extra_compile_args.append("-DUNICODE")

ext_modules = [
    CMakeExtension(
        "_nod",
        ["_nod.pyx", "py-nod/nod_wrap_util.cxx"],
        cmake_options={
            "dir": nod_submodule,
            "targets": {
                "nod": "lib",
                "logvisor": "logvisor",
            },
        },
        language='c++',
        extra_compile_args=extra_compile_args,
        extra_objects=[
        ],
    )
]


def create_extension(template, kwds):
    """"""
    kwds["cmake_options"] = template.cmake_options
    return default_create_extension(template, kwds)

cythonized_ext_modules = cythonize(
    ext_modules,
    include_path=custom_include_paths,
    create_extension=create_extension,
)

for ext_module in cythonized_ext_modules:
    ext_module.include_dirs = custom_include_paths

setup(
    name='nod',
    version="0.1.3",
    author='Henrique Gemignani',
    url='https://github.com/henriquegemignani/py-nod',
    description='Python bindings for the nod library.',
    packages=["nod"],
    scripts=[
    ],
    package_data={
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3 :: Only',
    ],
    install_requires=[],
    setup_requires=[
        'Cython',
    ],
    cmdclass={
        'build_ext': CMakeBuild,
    },
    ext_modules=cythonized_ext_modules,
)



from enum import Enum
from typing import Tuple, Optional, Callable

from libcpp cimport bool as c_bool
from libcpp.string cimport string
from libcpp.memory cimport unique_ptr

from nod_wrap cimport ExtractionContext as c_ExtractionContext, \
    createProgressCallbackFunction, DiscBase as c_DiscBase, \
    OpenDiscFromImage, SystemStringView, SystemUTF8View, SystemString, \
    DiscBuilderGCN as c_DiscBuilderGCN, createFProgressFunction, EBuildResult,\
    EBuildResult_Success, EBuildResult_Failed, EBuildResult_DiskFull

cdef str _system_string_to_str(SystemString path):
    return SystemUTF8View(path).utf8_str().decode("utf-8")

cdef SystemString _str_to_system_string(str path):
    return SystemStringView(path.encode("utf-8")).sys_str()

ProgressCallback = Callable[[float, str, int], None]

cdef void invoke_callback_function(object callback, const string& a, float progress):
    callback(a.decode("utf-8"), progress)

cdef void invoke_fprogress_function(object callback, float totalProg, const SystemString& fileName, size_t fileBytesXfered):
    callback(totalProg, _system_string_to_str(fileName), fileBytesXfered)

cdef class ExtractionContext:
    cdef c_ExtractionContext c_context

    def __cinit__(self):
        self.c_context = c_ExtractionContext()

    @property
    def force(self):
        return self.c_context.force

    @force.setter
    def force(self, value):
        self.c_context.force = value

    def set_progress_callback(self, callback):
        self.c_context.progressCB = createProgressCallbackFunction(callback, invoke_callback_function)

cdef class Partition:
    cdef c_DiscBase.IPartition*c_partition

    @staticmethod
    cdef create(c_DiscBase.IPartition* c_partition):
        partition = Partition()
        partition.c_partition = c_partition
        return partition

    def extract_to_directory(self, path: str, context: ExtractionContext) -> bool:
        return self.c_partition.extractToDirectory(
            _str_to_system_string(path),
            context.c_context
        )

cdef class DiscBase:
    cdef unique_ptr[c_DiscBase] c_disc

    def get_data_partition(self) -> Optional[Partition]:
        cdef c_DiscBase.IPartition*partition = self.c_disc.get().getDataPartition()
        if partition:
            return Partition.create(partition)
        else:
            return None

class BuildResult(Enum):
    Success = 1
    Failed = 2
    DiskFull = 3


cdef object convert_e_build_result(EBuildResult result):
    if result == EBuildResult_Success:
        return BuildResult.Success
    elif result == EBuildResult_Failed:
        return BuildResult.Failed
    elif result == EBuildResult_DiskFull:
        return BuildResult.DiskFull
    else:
        raise ValueError("Unknown EBuildResult")

cdef class DiscBuilderGCN:
    cdef c_DiscBuilderGCN* c_builder

    def __init__(self, out_path: str, progress_callback: ProgressCallback):
        pass

    def __cinit__(self, out_path: str, progress_callback: ProgressCallback):
        self.c_builder = new c_DiscBuilderGCN(_str_to_system_string(out_path).c_str(),
                                              createFProgressFunction(progress_callback, invoke_fprogress_function))

    def __dealloc__(self):
        del self.c_builder

    def build_from_directory(self, directory_in: str) -> BuildResult:
        return convert_e_build_result(self.c_builder.buildFromDirectory(_str_to_system_string(directory_in).c_str()))

    @staticmethod
    def calculate_total_size_required(directory_in: str) -> int:
        return c_DiscBuilderGCN.CalculateTotalSizeRequired(_str_to_system_string(directory_in).c_str())

def open_disc_from_image(path: str) -> Optional[Tuple[DiscBase, bool]]:
    disc = DiscBase()
    cdef c_bool is_wii = True
    disc.c_disc = OpenDiscFromImage(_str_to_system_string(path).c_str(), is_wii)

    if disc.c_disc:
        return disc, is_wii
    else:
        return None



import argparse

import os

import nod


class Commands:
    def __init__(self, args):
        self.args = args

    def fprogress_callback(self, progress: float, name: str, bytes: int):
        if self.args.verbose:
            print("\r" + " " * 100, end="")
            print("\r{:.0%} {} {} B".format(progress, name, bytes), flush=True)

    def extract(self):
        args = self.args

        result = nod.open_disc_from_image(args.image_in)
        if not result:
            if args.verbose:
                print("Could not open disc from '{}'.".format(args.image_in))
            raise SystemExit(1)

        disc, is_wii = result
        data_partition = disc.get_data_partition()
        if not data_partition:
            if args.verbose:
                print("Could not find a data partition in the disc.")
            raise SystemExit(2)

        def progress_callback(path, progress):
            if args.verbose:
                print("Extraction {:.0%} Complete; Current node: {}".format(progress, path))

        context = nod.ExtractionContext()
        context.set_progress_callback(progress_callback)

        if not data_partition.extract_to_directory(args.directory_out, context):
            if args.verbose:
                print("Could not extract to '{}'".format(args.directory_out))

    def makegcn(self):
        filesystem_root = self.args.filesystem_root

        if not os.path.isdir(filesystem_root):
            print("Error, '{}' is not a directory.".format(filesystem_root))
            raise SystemExit(1)

        if nod.DiscBuilderGCN.calculate_total_size_required(filesystem_root) == -1:
            print("Image built with given directory would pass the maximum size.")
            raise SystemExit(2)

        disc_builder = nod.DiscBuilderGCN(self.args.image_out, self.fprogress_callback)
        ret = disc_builder.build_from_directory(filesystem_root)

        if self.args.verbose:
            print()

        if ret != nod.BuildResult.Success:
            print("Failure building the image: code {}".format(ret))
            raise SystemExit(3)


    def execute(self):
        getattr(self, self.args.command)()


def create_parsers():
    parser = argparse.ArgumentParser()
    sub_parsers = parser.add_subparsers(dest="command")

    # Extract
    extract_parser = sub_parsers.add_parser(
        "extract",
        help="Extract an iso"
    )
    extract_parser.add_argument("image_in", type=str)
    extract_parser.add_argument("directory_out", type=str)
    extract_parser.add_argument("-v", "--verbose", action="store_true")

    # Make GCN
    make_gcn_parser = sub_parsers.add_parser(
        "makegcn",
        help="Create a Nintendo GameCube ISO from files."
    )
    make_gcn_parser.add_argument("filesystem_root", type=str)
    make_gcn_parser.add_argument("image_out", type=str)
    make_gcn_parser.add_argument("-v", "--verbose", action="store_true")

    return parser


def parse_args(parser):
    args = parser.parse_args()
    if args.command is None:
        parser.print_help()
        raise SystemExit(1)

    Commands(args).execute()

if __name__ == "__main__":
    parse_args(create_parsers())
